// Copyright (C) 2007 CEREA
//     Author(s): Meryem Ahmed de Biasi
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/

#ifndef COMMON_FILE_COMMON_CXX

#include "Common.hxx"

namespace Polyphemus
{

  //! Parse the arguments of a preprocessing program.
  /*!
    \param argc number of system arguments.
    \param argv table of system arguments.
    \param config_file (output) configuration file for the program.
    \param default_name (optional) default name of the configuration file.
    Default: "".
   */
  void parse_argument(int argc, char** argv, string& config_file,
		      string default_name)
  {
    string msg;
    if (default_name != "")
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [configuration file]\n";
	msg += string("  ") + argv[0] + "\n\n";
	msg += "Arguments:\n";
	msg += "  [configuration file] (optional): configuration";
	msg += " file. Default: "+ default_name + "\n";
      }
    else
       {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [configuration file]\n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] : main configuration file.\n";
       }
    if (argc > 2 || (default_name == "" && argc < 2))
	throw msg;

    if (argc == 1)
      config_file = default_name;
    else
      config_file = argv[1];

  }


  //! Parse the arguments of a preprocessing program.
  /*!
    \param argc number of system arguments.
    \param argv table of system arguments.
    \param first_file (output) main configuration file for the program.
    \param second_file (output) secondary configuration file for the program.
    \param default_name (optional) default name of the main configuration
    file. Default: "".
  */
  void parse_argument(int argc, char** argv, string& first_file,
		      string& second_file, string default_name)
  {
    string msg;
    if (default_name != "")
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += "[secondary config file] \n";
	msg += string("  ") + argv[0] + " [main configuration file] \n";
	msg += string("  ") + argv[0] + "\n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] (optional): main configuration";
	msg += " file. Default: "+ default_name + "\n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file.\n";
      }
    else
       {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += "[secondary config file] \n";
	msg += string("  ") + argv[0] + " [main configuration file] \n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] : main configuration file.\n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file. \n";
       }
    if (argc > 3 || (default_name == "" && argc < 2))
	throw msg;

    if (argc == 1)
      first_file = default_name;
    else
      first_file = argv[1];

    if (argc < 3)
      second_file = "";
    else
      second_file = argv[2];
  }
  

  //! Parse the arguments of a preprocessing program.
  /*!
    \param argc number of system arguments.
    \param argv table of system arguments.
    \param first_file (output) main configuration file for the program.
    \param second_file (output) secondary configuration file for the program.
    \param date (output) beginning date for the program.
    \param default_name (optional) default name of the main configuration
    file. Default: "".
  */
  void parse_argument(int argc, char** argv, string& first_file,
		      string& second_file, Date& date,
		      string default_name)
  {
    string msg;
    if (default_name != "")
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file] ";
	msg += "[secondary config file] [date]\n";
	msg += string("  ") + argv[0] + " [main configuration file] [date]\n";
	msg += string("  ") + argv[0] + " [date] \n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] (optional): main configuration";
	msg += " file. Default: "+ default_name + "\n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file.\n";
	msg += "  [date]: date in any valid format. \n";
      }
    else
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file] ";
	msg += "[secondary config file] [date]\n";
	msg += string("  ") + argv[0] + " [main configuration file]"
	  + " [date]\n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] : main configuration file.\n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file. \n";
	msg += "  [date]: date in any valid format. \n";
      }
    if (argc > 4 || !is_date(argv[argc - 1]) ||
	(default_name == "" && argc < 3) ||
	(default_name != "" && argc < 2))
      throw msg;

    if (argc < 3)
      first_file = default_name;
    else
      first_file = string(argv[1]);

    if (argc < 4)
      second_file = "";
    else
      second_file = string(argv[2]);
    
    if (!is_date(argv[argc - 1]))
      throw "Date " + string(argv[argc - 1]) + "is not in a valid format.";
    else
      date = string(argv[argc - 1]);
  }

  //! Parse the arguments of a preprocessing program.
  /*!
    \param argc number of system arguments.
    \param argv table of system arguments.
    \param first_file (output) main configuration file for the program.
    \param second_file (output) secondary configuration file for the program.
    \param date_beg (output) beginning date for the program.
    \param date_end (output) end date for the program.
    \param default_name (optional) default name of the main configuration
    file. Default: "".
   */
  void parse_argument(int argc, char** argv, string& first_file,
		   string& second_file, Date& date_beg, Date& date_end,
		   string default_name)
  {
    string msg;
    if (default_name != "")
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [secondary config file] [first date]";
	msg += " [second date/interval]\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [first date] [second date/interval] \n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [secondary config file] [first date] \n";
	msg += string("  ") + argv[0] + " [first date]";
	msg += " [second date/interval] \n";
	msg += string("  ") + argv[0] + " [first date]  \n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] (optional): main configuration";
	msg += " file. Default: "+ default_name + "\n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file. Default: \"\".\n";
	msg += "  [first date]: beginning date in any valid format. \n";
	msg += "  [second date]: end date in any valid format. \n";
	msg += "  [interval] (optional): Interval in format NdMh or Nd-Mh";
	msg += " or Nd or Mh where N is the number of days and M the number";
	msg += " of hours. Default: 1d. \n\n";
	msg += "Note:\n";
	msg += "The end date, whether it is given directly or computed by";
	msg += " adding the time interval to the beginning date, is always";
	msg += " considered as excluded. \n";
	
      }
    else
      {
	msg =  "Usage:\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [secondary config file] [first date]";
	msg += " [second date/interval]\n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [first date] [second date/interval] \n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [secondary config file] [first date] \n";
	msg += string("  ") + argv[0] + " [main configuration file]";
	msg += " [first date] \n\n";
	msg += "Arguments:\n";
	msg += "  [main configuration file] : main configuration file. \n";
	msg += "  [secondary configuration file] (optional): secondary";
	msg += " configuration file. Default: \"\".\n";
	msg += "  [first date]: beginning date in any valid format. \n";
	msg += "  [second date]: end date in any valid format. \n";
	msg += "  [interval] (optional): Interval in format NdMh or Nd-Mh";
	msg += " or Nd or Mh where N is the number of days and M the number";
	msg += " of hours. Default: 1d. \n\n";
	msg += "Note:\n";
	msg += "The end date, whether it is given directly or computed by";
	msg += " adding the time interval to the beginning date, is always";
	msg += " considered as excluded. \n";
      }

    if (argc > 5 || (default_name == "" && argc < 3) ||
	(default_name != "" && argc < 2) || (!is_date(argv[argc - 1]) &&
					     !is_delta(argv[argc - 1])))
      throw msg;

    if (is_date(argv[1]))
      {
	first_file = default_name;
	second_file = "";
	date_beg = string(argv[1]);
	if (argc == 2)
	  {
	    date_end = date_beg;
	    date_end.AddDays(1);
	  }
	else
	  {
	    if (is_date(argv[2]))
	      date_end = string(argv[2]);
	    else
	      date_end = convert_delta(string(argv[2]), date_beg);
	  }
      }
    else
      {
	first_file = string(argv[1]);
	if (is_date(argv[2]))
	  {
	    date_beg = string(argv[2]);
	    if (argc == 3)
	      {
		date_end = date_beg;
		date_end.AddDays(1);
	      }
	    else
	      {
		if (is_date(argv[3]))
		  date_end = string(argv[3]);
		else
		  date_end = convert_delta(string(argv[3]), date_beg);
	      }
	  }
	else
	  {
	    second_file = string(argv[2]);
	    date_beg = string(argv[3]);
	    if (argc == 4)
	      {
		date_end = date_beg;
		date_end.AddDays(1);
	      }
	    else
	      {
		if (is_date(argv[4]))
		  date_end = string(argv[4]);
		else
		  date_end = convert_delta(string(argv[4]), date_beg);
	      }
	  }
      }
  }


  //! Reads the beginning date of a MM5 file.
  /*! Note that the beginning date in the BigHeader is not necessarily the
    date we need.
    Indeed, data may have been generated for a long period of time but stored
    in several files. In that case, the date given in the BigHeader is the
    beginning date for the first file.
    \param FileName name of the MM5 file.
    \return The beginning  date.
   */
  Date read_date_MM5(string FileName)
  {
    ifstream file(FileName.c_str());
    FormatMM5 InputMeteo;
    MM5SubHeader SubHeader;
    
    while(InputMeteo.ReadFlag(file) != 1)
      {
	Array<int, 2> BHI;
	Array<float, 2> BHR;
	Array<string, 2> BHIC;
	Array<string, 2> BHRC;
	InputMeteo.ReadBigHeader(file, BHI, BHR, BHIC, BHRC);
      }
    InputMeteo.ReadSubHeader(file, SubHeader);
    string current_date = SubHeader.GetCurrentDate();
    current_date = current_date.substr(0, 19);
    current_date = find_replace(current_date, ":", "-");
    Date date(current_date);
    return date;
  }


  //! Computes the end date from the beginning date and the time interval.
  /*!
    \param delta time interval, in the format "NdMh" or "Nd-Mh" where N is the
    number of days (integer) and M the number of hours (integer). If N is
    zero, the time interval can be expressed as "Mh".
    \param date_beg beginning date.
    \return The end date.
   */
  Date convert_delta(string delta, Date date)
  {
    Date date_end = date;
    
    if(!is_delta(delta))
      throw "Wrong time period " + delta + string(".");
    else
      {
	vector<string> period = split(delta, "dh-_");
	int days = 0, hours = 0;
	if (period.size() == 2)
	  {
	    to_num(period[0], days);
	    to_num(period[1], hours);
	  }
	else if(delta.find('h', 0) != string::npos)
	  to_num(period[0], hours);
	else if(delta.find('d', 0) != string::npos)
	  to_num(period[0], days);
	date_end.AddDays(days);
	date_end.AddHours(hours);
	return date_end;
      }
  }


  //! Compute the number of iterations from two dates and a time step.
  /*! The number of iterations is computed to go from the beginning date
    included to the end date excluded.
    If the time between the beginning and end date of the preprocessing is not
    a multiple of the time step, an exception is raised.
    \param date_beg beginning date of the preprocessing.
    \param date_end end date of the preprocessing.
    \param Delta_t time step in hours.
    \return The number of iteration for the preprocessing (integer).
   */
  template<typename T>
  int compute_Nt(Date date_beg, Date date_end, T Delta_t)
    {
      int distance = int(date_end.GetSecondsFrom(date_beg));
      int Delta_t_sec =  int(Delta_t * 3600);
      if (distance % Delta_t_sec != 0)
	throw string("The time between the beginning and end date must be a ")
	  + "multiple of the time step (" + to_str(Delta_t) + " hours).";
      int Nt = distance / Delta_t_sec;
      return Nt;
    }


} //namespace Polyphemus

#define COMMON_FILE_COMMON_CXX
#endif
