// Copyright (C) 2005-2007 CEREA
//     Author(s): Irne Korsakissok, Vivien Mallet
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/

// This file is part of a Gaussian puff model for Polyphemus.


#ifndef POLYPHEMUS_FILE_MODELS_GAUSSIANPUFF_CXX


#include "BriggsFormula.hxx"
#include "GaussianPuff.hxx"


namespace Polyphemus
{


  //! Default constructor.
  /*!
    Builds the model. Nothing else is performed.
  */
  template<class T>
  GaussianPuff<T>::GaussianPuff():
    BaseModel<T>(), Npuff(0), cos_angle_(0.), sin_angle_(0.)
  {
  }


  //! Main constructor.
  /*!
    \param config_file configuration filename.
  */
  template<class T>
  GaussianPuff<T>::GaussianPuff(string config_file):
    BaseModel<T>(config_file), Npuff(0), cos_angle_(0.), sin_angle_(0.)
  {
  }


  //! Model initialization.
  /*! It reads the configuration.
   */
  template<class T>
  void GaussianPuff<T>::Init()
  {
    BaseModel<T>::Init();

    /*** Domain ***/

    this->config.SetSection("[domain]");

    // Land category ("rural" or "urban").
    this->config.PeekValue("Land_category", land_category);
    if (land_category != "rural" && land_category != "urban")
      throw string("Land category should be \"rural\" or \"urban\", but \"")
	+ land_category + "\" was provided in section \"[domain]\".";

    // Time (Night or Day).
    this->config.PeekValue("Time", day_night);
    if (day_night != "day" && day_night != "night")
      throw string("Time should be \"day\" or \"night\", but \"")
	+ day_night + "\" was provided in section \"[domain]\".";

    /*** Gaussian options ***/

    this->config.SetSection("[gaussian]");

    this->config.PeekValue("With_radioactive_decay",
			   this->option_process["with_radioactive_decay"]);
    this->config.PeekValue("With_biological_decay",
			   this->option_process["with_biological_decay"]);
    this->config.PeekValue("With_scavenging",
			   this->option_process["with_scavenging"]);
    this->config.PeekValue("With_dry_deposition",
			   this->option_process["with_deposition"]);

    // Parameterization to compute standard deviations ("Briggs", "Doury").
    this->config.PeekValue("Sigma_parameterization", sigma_formula);
    if (sigma_formula != "Briggs" && sigma_formula != "Doury")
      throw string("Sigma formula should be \"Briggs\" or \"Doury\", but \"")
	+ sigma_formula + "\" was provided in section \"[gaussian]\".";

    // Gets the deposition model.
    if (this->option_process["with_deposition"])
      {
	this->config.SetSection("[deposition]");
	this->config.PeekValue("Deposition_model", deposition_model);

	if (deposition_model == "Chamberlain")
	  this->config.PeekValue("Nchamberlain", Nchamberlain);
	else if (deposition_model != "Overcamp")
	  throw string("Deposition model should be \"Chamberlain\" or")
	    + " \"Overcamp\" but \"" + deposition_model + "\" was provided "
	    + "in section \"[deposition]\".";
      }
    else
      Nchamberlain = 0;

    /*** Species data ***/

    ConfigStream species_stream(this->file_species);

    // Reads the species half-lives in the species file.
    if (this->option_process["with_radioactive_decay"])
      {
	half_life_time.resize(this->Ns);

	species_stream.SetSection("[half_life]");
	for (int i = 0; i < this->Ns; i++)
	  species_stream.PeekValue(this->species_list[i], half_life_time(i));
	
	// Conversion from days to seconds.
	T factor = 24. * 3600.;
	for (int i = 0; i < this->Ns; i++)
	  half_life_time(i) *= factor;
     }

    // Reads the species biological half-lives in the species file.
    if (this->option_process["with_biological_decay"])
      {
	biological_half_life_time.resize(this->Ns);

	T day_value, night_value;
	for (int i = 0; i < this->Ns; i++)
	  {
	    species_stream.SetSection("[half_life_time]");
	    species_stream.Find(this->species_list[i]);
	    species_stream.GetNumber(day_value);
	    species_stream.GetNumber(night_value);
	    if (day_night == "day")
	      biological_half_life_time(i) = day_value;
	    else
	      biological_half_life_time(i) = night_value;
	  }
      }
  }


  //! Method called at each time step to initialize the model.
  /*!
    \note Empty method.
   */
  template<class T>
  void GaussianPuff<T>::InitStep()
  {
    BaseModel<T>::InitStep();
  }


  //! Puff source initialization.
  /*! It sets all sources from a text file. Each source is described in a
    dedicated section "[source]" in which one finds the following entries:
    <ul>
    <li> Abscissa: abscissa (m),
    <li> Ordinate: ordinate (m),
    <li> Altitude: height (m),
    <li> Species_name: species name.
    <li> Release_time: the release time (seconds),
    <li> Quantity: the released quantity (mass unit),
    </ul>
    \param source_file file that describes the sources.
  */
  template<class T>
  void GaussianPuff<T>::InitPuffSource()
  {
    // File containing puffs data.
    string file_puff;
    this->config.SetSection("[gaussian]");
    this->config.PeekValue("File_puff", file_puff);
    ConfigStream puff_stream(file_puff);

    // Puffs definitions.
    T time_puff, quantity, abscissa, ordinate, height;
    string line, species_name;

    PuffList.clear();
    Npuff = 0;
    while (!puff_stream.IsEmpty())
      {
	line = puff_stream.GetLine();
	if (split(line)[0] == "[source]")
	  {
	    puff_stream.PeekValue("Abscissa", abscissa);
	    puff_stream.PeekValue("Ordinate", ordinate);
	    puff_stream.PeekValue("Altitude", height);
	    puff_stream.PeekValue("Release_time", time_puff);
	    puff_stream.PeekValue("Quantity", quantity);
	    puff_stream.PeekValue("Species_name", species_name);

	    PuffList.push_back(Puff<T>(time_puff, quantity,
				       abscissa, ordinate, height,
				       this->GetSpeciesIndex(species_name)));
	    Npuff++;
	  }
      }

    // Sets the half-lives of the puff species.
    if (this->option_process["with_radioactive_decay"])
      for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	   iter != PuffList.end(); iter++)
	iter->SetHalfTime(half_life_time(iter->GetSpeciesIndex()));

    // Sets the biological half-lives of the puff species.
    if (this->option_process["with_biological_decay"])
      for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	   iter != PuffList.end(); iter++)
	{
	  int s = iter->GetSpeciesIndex();
	  iter->SetBioHalfTime(biological_half_life_time(s));
	}
  }


  //! Addition of a puff source.
  /*! It adds a new puff source with the given parameters.
    \param time_puff: the release time (seconds),
    \param quantity: the released quantity (mass unit),
    \param abscissa: abscissa (m),
    \param ordinate: ordinate (m),
    \param height: height (m),
    \param species_name: species name.
  */
  template<class T>
  void GaussianPuff<T>::AddPuffSource(T time_puff, T quantity,T abscissa,
				       T ordinate, T height,
				       string species_name)
  {
    PuffList.push_back(Puff<T>(time_puff, quantity,
			       abscissa, ordinate, height,
			       this->GetSpeciesIndex(species_name)));
    Npuff++;

  }

  //! Returns list of puffs.
  /*!
    \return the list of puffs.
   */
  template<class T>
  const list<Puff<T> >& GaussianPuff<T>::GetPuffList() const
  {
    return PuffList;
  }

  //! Sets the list of puffs.
  /*!
    \param pufflist the list of puffs.
  */
  template<class T>
  void GaussianPuff<T>::SetPuffList(list<Puff<T> > pufflist)
  {
    PuffList.clear();
    PuffList = pufflist;
  }


  //! Initializes puffs position and resets time.
  /*! It reinitializes puffs position, date, time, and Nchamberlain_tot.
  */
  template<class T>
  void GaussianPuff<T>::InitPosition()
  {
    /*** Initialization of puffs positions ***/

    for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	 iter != PuffList.end(); iter++)
      iter->InitPuff();

    // Initializes time, step, and Nchamberlain_tot.
    this->current_time = 0.;
    this->current_date = this->Date_min;
    this->step = 0;
    Nchamberlain_tot = Nchamberlain;
  }


  //! Initializes meteorological conditions.
  /*! It sets the meteorological data from a configuration file.  The
    situation is described in a dedicated section "[situation]" in which one
    finds the following entries:
    <ul>
    <li> Temperature: the temperature of ambient air (Celsius degrees).
    <li> Wind_angle: the angle between positive x-axis and wind,
    counterclockwise (degrees).
    <li> Wind: the wind velocity (m/s).
    <li> Stability: stability class in [A, F].
    </ul>
    \param meteo ConfigStream instance through which all entries (Temperature,
    Wind_angle, ...) may be read to set the meteorological situation.
    \param show_meteo indicates whether the meteorological data is to be
    displayed on screen.
  */
  template<class T>
  void GaussianPuff<T>::InitMeteo(ConfigStream& meteo, bool show_meteo)
  {

    /*** Main meteorological variables ***/

    meteo.PeekValue("Temperature", temperature_);
    meteo.PeekValue("Wind_angle", wind_angle_);
    meteo.PeekValue("Wind", wind_);
    meteo.PeekValue("Inversion_height", inversion_height_);
    meteo.PeekValue("Stability", stability_class_);

    if (show_meteo)
      cout << "\t" << temperature_
	   << "\t\t" << wind_angle_
	   << "\t\t" << wind_
	   << "\t\t" << stability_class_ << endl;

    temperature_ = temperature_ + 273.15;
    T wind_angle_rad = wind_angle_ / 180. * 3.14159265358979323846264;
    cos_angle_ = cos(wind_angle_rad);
    sin_angle_ = sin(wind_angle_rad);

    stability_class_ = upper_case(stability_class_);
    if (stability_class_ == "A")
      stability_ = 0;
    else if (stability_class_ == "B")
      stability_ = 1;
    else if (stability_class_ == "C")
      stability_ = 2;
    else if (stability_class_ == "D")
      stability_ = 3;
    else if (stability_class_ == "E")
      stability_ = 4;
    else if (stability_class_ == "F")
      stability_ = 5;
    else
      throw string("Stability class should be in [A, F], but \"")
	+ stability_class_ + "\" was provided.";

    /*** Scavenging coefficients ***/

    if (this->option_process["with_scavenging"])
      {
	scavenging_coefficient.resize(this->Ns);

	meteo.Find("Scavenging_coefficient");
	vector<string> scav_coefficient =  split(meteo.GetLine());
	vector<string>::iterator iter;
	for (int i = 0; i < this->Ns; i++)
	  {
	    iter = find(scav_coefficient.begin(), scav_coefficient.end(),
			this->species_list[i]);
	    if (iter++ == scav_coefficient.end() ||
		iter == scav_coefficient.end())
	      throw string("Unable to find scavenging coefficient for")
		+ string(" species \"") + this->species_list[i] + "\".";
	    scavenging_coefficient(i) = to_num<T>(*iter);
	  }

	// Sets the scavenging coefficient for each puff species.
	for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	     iter != PuffList.end(); iter++)
	  {
	    int s = iter->GetSpeciesIndex();
	    iter->SetScavengingCoefficient(scavenging_coefficient(s));
	  }
      }

    /*** Deposition velocities ***/

    if (this->option_process["with_deposition"])
      {
	deposition_velocity.resize(this->Ns);

	meteo.Find("Deposition_velocity");
	vector<string> dep_velocity = split(meteo.GetLine());
	vector<string>::iterator iter;
	for (int  i = 0; i < this->Ns; i++)
	  {
	    iter = find(dep_velocity.begin(), dep_velocity.end(),
			this->species_list[i]);
	    if (iter++ == dep_velocity.end() ||
		iter == dep_velocity.end())
	      throw string("Unable to find deposition velocity for")
		+ string(" species \"") + this->species_list[i] + "\".";
	    deposition_velocity(i) = to_num<T>(*iter);
	  }

	// Sets the deposition velocity for each puff species.
	for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	     iter != PuffList.end(); iter++)
	  {
	    int s = iter->GetSpeciesIndex();
	    iter->SetDepositionVelocity(deposition_velocity(s));
	  }
      }
  }


  //! Sets values of meteorological parameters.
  /*! It sets the meteorological data.
    \param temperature: the temperature of ambient air (Celsius degrees).
    \param wind_angle: the angle between positive x-axis and wind,
    counterclockwise (degrees).
    \param wind: the wind velocity (m/s).
    \param stability: stability class in [A, F].
  */
  template<class T>
  void GaussianPuff<T>::SetMeteo(T temperature, T wind_angle, T wind,
				  string stability_class,
				  T inversion_height = 0.)
  {
    temperature_ = temperature;
    temperature_ = temperature_ + 273.15;
    wind_angle_ = wind_angle;
    T wind_angle_rad = wind_angle_ / 180. * 3.14159265358979323846264;
    cos_angle_ = cos(wind_angle_rad);
    sin_angle_ = sin(wind_angle_rad);
    wind_ = wind;
    inversion_height_ = inversion_height;

    stability_class_ = upper_case(stability_class);
    if (stability_class_ == "A")
      stability_ = 0;
    else if (stability_class_ == "B")
      stability_ = 1;
    else if (stability_class_ == "C")
      stability_ = 2;
    else if (stability_class_ == "D")
      stability_ = 3;
    else if (stability_class_ == "E")
      stability_ = 4;
    else if (stability_class_ == "F")
      stability_ = 5;
    else
      throw string("Stability class should be in [A, F], but \"")
	+ stability_class_ + "\" was provided.";
  }


  //! Sets time of the day (night or day).
  template<class T>
  void GaussianPuff<T>::SetDayNight(bool isday)
  {
    if (isday == 0)
      day_night = "night";
    else
      day_night = "day";
  }


  //! Computes the new position of the puff center after advection.
  template<class T>
  void GaussianPuff<T>::Advection()
  {
    for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	 iter != PuffList.end(); iter++)
      if (this->current_time >= iter->GetReleaseTime())
	iter->Advection(wind_, cos_angle_, sin_angle_, this->Delta_t);
  }


  //!  Computes the horizontal and vertical standard deviations.
  template<class T>
  void GaussianPuff<T>::Diffusion()
  {
    T sigma_x, sigma_y, sigma_z;
    bool rural = land_category == "rural";
    bool briggs = sigma_formula == "Briggs";
    bool day = day_night == "day";

    for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	 iter != PuffList.end(); iter++)
      if (this->current_time >= iter->GetReleaseTime())
	{
	  T distance_x = iter->GetDistance();
	  // Horizontal diffusion parameter.
	  if (briggs)
	    if (rural)
	      sigma_y =
		ComputeRuralPlumeHorizontalSigma(distance_x, stability_);
	    else
	      sigma_y =
		ComputeUrbanPlumeHorizontalSigma(distance_x, stability_);
	  else
	    {
	      T transfert_time = distance_x / wind_;
	      if (day)
		sigma_y =
		  ComputeDouryNormalPlumeHorizontalSigma(transfert_time);
	      else if (wind_ > 3.)
		sigma_y =
		  ComputeDouryNormalPlumeHorizontalSigma(transfert_time);
	      else
		sigma_y =
		  ComputeDouryLowPlumeHorizontalSigma(transfert_time);
	    }
	  sigma_x = sigma_y;
	  
	  // Vertical diffusion parameter.
	  if (briggs)
	    if (rural)
	      sigma_z =
		ComputeRuralPlumeVerticalSigma(distance_x, stability_);
	    else
	      sigma_z =
		ComputeUrbanPlumeVerticalSigma(distance_x, stability_);
	  else
	    {
	      T transfert_time = distance_x / wind_;
	      if (day)
		sigma_z =
		  ComputeDouryNormalPlumeVerticalSigma(transfert_time);
	      else if (wind_ > 3.)
		sigma_z =
		  ComputeDouryNormalPlumeVerticalSigma(transfert_time);
	      else
		sigma_z =
		  ComputeDouryLowPlumeVerticalSigma(transfert_time);
	    }
	  iter->SetSigma_x(sigma_x);
	  iter->SetSigma_y(sigma_y);
	  iter->SetSigma_z(sigma_z);
	}
  }


  //! Performs one step forward.
  template<class T>
  void GaussianPuff<T>::Forward()
  {
    Advection();
    Diffusion();

    int i, j, k;
    this->Concentration.SetZero();
    for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	 iter != PuffList.end(); iter++)
      if (this->current_time >= iter->GetReleaseTime())
	for (k = 0; k < this->Nz; k++)
	  for (j = 0; j < this->Ny; j++)
	    for (i = 0; i < this->Nx; i++)
	      this->Concentration(iter->GetSpeciesIndex(), k, j, i) +=
		iter->ComputeConcentration(this->GridX4D(i), this->GridY4D(j),
					   this->GridZ4D(k), wind_,
					   cos_angle_, sin_angle_,
					   stability_, inversion_height_,
					   land_category,
					   day_night, sigma_formula,
					   this->option_process
					   ["with_radioactive_decay"],
					   this->option_process
					   ["with_biological_decay"],
					   this->option_process
					   ["with_scavenging"],
					   this->option_process
					   ["with_deposition"],
					   deposition_model,
					   Nchamberlain_tot);
    Nchamberlain_tot += Nchamberlain;
    this->AddTime(this->Delta_t);
    this->step++;
  }


  //! Computes concentration at a given point.
  /*!
    \return The concentration at the point.
   */
  template<class T>
  T GaussianPuff<T>::GetConcentration(int species, T z, T y, T x)
  {
    T concentration = 0.;
    this->SubtractTime(this->Delta_t);
    for (typename list<Puff<T> >::iterator iter = PuffList.begin();
	 iter != PuffList.end(); iter++)
      if (this->current_time >= iter->GetReleaseTime()
	  && iter->GetSpeciesIndex() == species)
	concentration +=
	  iter->ComputeConcentration(x, y, z,
				     wind_,
				     cos_angle_, sin_angle_,
				     stability_, inversion_height_,
				     land_category,
				     day_night, sigma_formula,
				     this->option_process
				     ["with_radioactive_decay"],
				     this->option_process
				     ["with_biological_decay"],
				     this->option_process
				     ["with_scavenging"],
				     this->option_process
				     ["with_deposition"],
				     deposition_model,
				     Nchamberlain_tot);
    this->AddTime(this->Delta_t);
    return concentration;
  }


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODELS_GAUSSIANPUFF_CXX
#endif
